/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2003 by Ralf Baechle
 */
#ifndef __ASM_MACH_JAZZ_PARAM_H
#define __ASM_MACH_JAZZ_PARAM_H

/*
 * Jazz is currently using the internal 100Hz timer of the R4030
 */
#define HZ		100		/* Internal kernel timer frequency */

#endif /* __ASM_MACH_JAZZ_PARAM_H */
