/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2003 by Ralf Baechle
 */
/*
 * Changes:
 *  Yoichi Yuasa <yuasa@hh.iij4u.or.jp>
 *  - CLOCK_TICK_RATE is changed into 32768 from 6144000.
 */
#ifndef __ASM_MACH_VR41XX_TIMEX_H
#define __ASM_MACH_VR41XX_TIMEX_H

#define CLOCK_TICK_RATE		32768

#endif /* __ASM_MACH_VR41XX_TIMEX_H */
